@@ -13,11 +13,12 @@
 #endif
 
 #define SPG_MAX_FIELDS 256
-#define SPG_MILLISECONDS_PER_DAY 86400000000.0
+#define SPG_MICROSECONDS_PER_DAY_LL 86400000000ULL
+#define SPG_MICROSECONDS_PER_DAY 86400000000.0
 #define SPG_MINUTES_PER_DAY 1440.0
 #define SPG_SECONDS_PER_DAY 86400.0
 
-#define SPG_DT_ADD_USEC if (usec != 0) { dt = rb_funcall(dt, spg_id_op_plus, 1, rb_float_new(usec/SPG_MILLISECONDS_PER_DAY)); }
+#define SPG_DT_ADD_USEC if (usec != 0) { dt = rb_funcall(dt, spg_id_op_plus, 1, spg_id_Rational ? rb_funcall(rb_cObject, spg_id_Rational, 2, INT2NUM(usec), ULL2NUM(SPG_MICROSECONDS_PER_DAY_LL)) : rb_float_new(usec/SPG_MICROSECONDS_PER_DAY)); }
 
 #define SPG_NO_TZ 0
 #define SPG_DB_LOCAL 1
@@ -35,12 +36,28 @@
 #define SPG_YIELD_KMV_HASH 7
 #define SPG_YIELD_MKMV_HASH 8
 #define SPG_YIELD_MODEL 9
+#define SPG_YIELD_KV_HASH_GROUPS 10
+#define SPG_YIELD_MKV_HASH_GROUPS 11
+#define SPG_YIELD_KMV_HASH_GROUPS 12
+#define SPG_YIELD_MKMV_HASH_GROUPS 13
+
+struct spg_row_proc_info {
+  VALUE dataset;
+  VALUE block;
+  VALUE model;
+  VALUE colsyms[SPG_MAX_FIELDS];
+  VALUE colconvert[SPG_MAX_FIELDS];
+#if SPG_ENCODING
+  int enc_index;
+#endif
+};
 
 static VALUE spg_Sequel;
 static VALUE spg_Blob;
 static VALUE spg_BigDecimal;
 static VALUE spg_Date;
 static VALUE spg_SQLTime;
+static VALUE spg_PGError;
 
 static VALUE spg_sym_utc;
 static VALUE spg_sym_local;
@@ -48,10 +65,16 @@ static VALUE spg_sym_map;
 static VALUE spg_sym_first;
 static VALUE spg_sym_array;
 static VALUE spg_sym_hash;
+static VALUE spg_sym_hash_groups;
 static VALUE spg_sym_model;
 static VALUE spg_sym__sequel_pg_type;
 static VALUE spg_sym__sequel_pg_value;
 
+static VALUE spg_nan;
+static VALUE spg_pos_inf;
+static VALUE spg_neg_inf;
+
+static ID spg_id_Rational;
 static ID spg_id_new;
 static ID spg_id_local;
 static ID spg_id_year;
@@ -66,6 +89,8 @@ static ID spg_id_utc;
 static ID spg_id_utc_offset;
 static ID spg_id_localtime;
 static ID spg_id_new_offset;
+static ID spg_id_convert_infinite_timestamps;
+static ID spg_id_infinite_timestamp_value;
 
 static ID spg_id_call;
 static ID spg_id_get;
@@ -89,6 +114,115 @@ static int enc_get_index(VALUE val)
 }
 #endif
 
+static VALUE read_array(int *index, char *c_pg_array_string, int array_string_length, char *word, VALUE converter)
+{
+  int word_index = 0;
+
+  /* The current character in the input string. */
+  char c;
+
+  /*  0: Currently outside a quoted string, current word never quoted
+   *  1: Currently inside a quoted string
+   * -1: Currently outside a quoted string, current word previously quoted */
+  int openQuote = 0;
+
+  /* Inside quoted input means the next character should be treated literally,
+   * instead of being treated as a metacharacter.
+   * Outside of quoted input, means that the word shouldn't be pushed to the array,
+   * used when the last entry was a subarray (which adds to the array itself). */
+  int escapeNext = 0;
+
+  VALUE array = rb_ary_new();
+
+  /* Special case the empty array, so it doesn't need to be handled manually inside
+   * the loop. */
+  if(((*index) < array_string_length) && c_pg_array_string[(*index)] == '}') 
+  {
+    return array;
+  }
+
+  for(;(*index) < array_string_length; ++(*index))
+  {
+    c = c_pg_array_string[*index];
+    if(openQuote < 1)
+    {
+      if(c == ',' || c == '}')
+      {
+        if(!escapeNext)
+        {
+          if(openQuote == 0 && word_index == 4 && !strncmp(word, "NULL", word_index))
+          {
+            rb_ary_push(array, Qnil);
+          }
+          else if (RTEST(converter))
+          {
+            rb_ary_push(array, rb_funcall(converter, spg_id_call, 1, rb_str_new(word, word_index)));
+          }
+          else
+          {
+            rb_ary_push(array, rb_str_new(word, word_index));
+          }
+        }
+        if(c == '}')
+        {
+          return array;
+        }
+        escapeNext = 0;
+        openQuote = 0;
+        word_index = 0;
+      }
+      else if(c == '"')
+      {
+        openQuote = 1;
+      }
+      else if(c == '{')
+      {
+        (*index)++;
+        rb_ary_push(array, read_array(index, c_pg_array_string, array_string_length, word, converter));
+        escapeNext = 1;
+      }
+      else
+      {
+        word[word_index] = c;
+        word_index++;
+      }
+    }
+    else if (escapeNext) {
+      word[word_index] = c;
+      word_index++;
+      escapeNext = 0;
+    }
+    else if (c == '\\')
+    {
+      escapeNext = 1;
+    }
+    else if (c == '"')
+    {
+      openQuote = -1;
+    }
+    else
+    {
+      word[word_index] = c;
+      word_index++;
+    }
+  }
+
+  return array;
+}
+
+static VALUE parse_pg_array(VALUE self, VALUE pg_array_string, VALUE converter) {
+
+  /* convert to c-string, create additional ruby string buffer of
+   * the same length, as that will be the worst case. */
+  char *c_pg_array_string = StringValueCStr(pg_array_string);
+  int array_string_length = RSTRING_LEN(pg_array_string);
+  VALUE buf = rb_str_buf_new(array_string_length);
+  char *word = RSTRING_PTR(buf);
+  int index = 1;
+
+  return read_array(&index, c_pg_array_string, array_string_length, word, converter);
+}
+
 static VALUE spg_time(const char *s) {
   VALUE now;
   int hour, minute, second, tokens;
@@ -117,6 +251,17 @@ static VALUE spg_date(const char *s) {
   return rb_funcall(spg_Date, spg_id_new, 3, INT2NUM(year), INT2NUM(month), INT2NUM(day));
 }
 
+static VALUE spg_timestamp_error(const char *s, VALUE self) {
+  VALUE db;
+  db = rb_funcall(self, spg_id_db, 0);
+  if(RTEST(rb_funcall(db, spg_id_convert_infinite_timestamps, 0))) {
+    if((strcmp(s, "infinity") == 0) || (strcmp(s, "-infinity") == 0)) {
+      return rb_funcall(db, spg_id_infinite_timestamp_value, 1, rb_tainted_str_new2(s));
+    }
+  }
+  rb_raise(rb_eArgError, "unexpected datetime format");
+}
+
 static VALUE spg_timestamp(const char *s, VALUE self) {
   VALUE dtc, dt, rtz;
   int tz = SPG_NO_TZ;
@@ -134,7 +279,7 @@ static VALUE spg_timestamp(const char *s, VALUE self) {
        	&usec_start, &usec, &usec_stop, 
 	&offset_sign, &offset_hour, &offset_minute);
     if(tokens < 7) {
-      rb_raise(rb_eArgError, "unexpected datetime format");
+      return spg_timestamp_error(s, self);
     }
     usec *= (int) pow(10, (6 - (usec_stop - usec_start)));
   } else {
@@ -146,7 +291,7 @@ static VALUE spg_timestamp(const char *s, VALUE self) {
       min = 0;
       sec = 0;
     } else if (tokens < 6) {
-      rb_raise(rb_eArgError, "unexpected datetime format");
+      return spg_timestamp_error(s, self);
     }
     usec = 0;
   }
@@ -289,7 +434,15 @@ static VALUE spg__col_value(VALUE self, PGresult *res, long i, long j, VALUE* co
         break;
       case 700: /* float */
       case 701:
-        rv = rb_float_new(rb_cstr_to_dbl(v, Qfalse));
+        if (strcmp("NaN", v) == 0) {
+          rv = spg_nan;
+        } else if (strcmp("Infinity", v) == 0) {
+          rv = spg_pos_inf;
+        } else if (strcmp("-Infinity", v) == 0) {
+          rv = spg_neg_inf;
+        } else {
+          rv = rb_float_new(rb_cstr_to_dbl(v, Qfalse));
+        }
         break;
       case 790: /* numeric */
       case 1700:
@@ -367,32 +520,13 @@ static VALUE spg__field_ids(VALUE v, VALUE *colsyms, long nfields) {
   return pg_columns;
 }
 
-static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
-  PGresult *res;
-  VALUE colsyms[SPG_MAX_FIELDS];
-  VALUE colconvert[SPG_MAX_FIELDS];
-  long ntuples;
-  long nfields;
+static void spg_set_column_info(VALUE self, PGresult *res, VALUE *colsyms, VALUE *colconvert) {
   long i;
   long j;
-  VALUE h;
+  long nfields;
   VALUE conv_procs = 0;
-  VALUE opts;
-  VALUE pg_type;
-  VALUE pg_value;
-  char type = SPG_YIELD_NORMAL;
 
-#ifdef SPG_ENCODING
-  int enc_index;
-  enc_index = enc_get_index(rres);
-#endif
-
-  Data_Get_Struct(rres, PGresult, res);
-  ntuples = PQntuples(res);
   nfields = PQnfields(res);
-  if (nfields > SPG_MAX_FIELDS) {
-    rb_raise(rb_eRangeError, "more than %d columns in query", SPG_MAX_FIELDS);
-  }
 
   for(j=0; j<nfields; j++) {
     colsyms[j] = rb_funcall(self, spg_id_output_identifier, 1, rb_str_new2(PQfname(res, j)));
@@ -427,6 +561,36 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
         break;
     }
   }
+}
+
+static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
+  PGresult *res;
+  VALUE colsyms[SPG_MAX_FIELDS];
+  VALUE colconvert[SPG_MAX_FIELDS];
+  long ntuples;
+  long nfields;
+  long i;
+  long j;
+  VALUE h;
+  VALUE opts;
+  VALUE pg_type;
+  VALUE pg_value;
+  char type = SPG_YIELD_NORMAL;
+
+#ifdef SPG_ENCODING
+  int enc_index;
+  enc_index = enc_get_index(rres);
+#endif
+
+  Data_Get_Struct(rres, PGresult, res);
+  ntuples = PQntuples(res);
+  nfields = PQnfields(res);
+  if (nfields > SPG_MAX_FIELDS) {
+    rb_raise(rb_eRangeError, "more than %d columns in query", SPG_MAX_FIELDS);
+  }
+
+  spg_set_column_info(self, res, colsyms, colconvert);
+
   rb_ivar_set(self, spg_id_columns, rb_ary_new4(nfields, colsyms));
 
   opts = rb_funcall(self, spg_id_opts, 0);
@@ -444,21 +608,21 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
         type = SPG_YIELD_FIRST;
       } else if (pg_type == spg_sym_array) {
         type = SPG_YIELD_ARRAY;
-      } else if (pg_type == spg_sym_hash && rb_type(pg_value) == T_ARRAY) {
+      } else if ((pg_type == spg_sym_hash || pg_type == spg_sym_hash_groups) && rb_type(pg_value) == T_ARRAY) {
         VALUE pg_value_key, pg_value_value;
         pg_value_key = rb_ary_entry(pg_value, 0);
         pg_value_value = rb_ary_entry(pg_value, 1);
         if (SYMBOL_P(pg_value_key)) {
           if (SYMBOL_P(pg_value_value)) {
-            type = SPG_YIELD_KV_HASH;
+            type = pg_type == spg_sym_hash_groups ? SPG_YIELD_KV_HASH_GROUPS : SPG_YIELD_KV_HASH;
           } else if (rb_type(pg_value_value) == T_ARRAY) {
-            type = SPG_YIELD_KMV_HASH;
+            type = pg_type == spg_sym_hash_groups ? SPG_YIELD_KMV_HASH_GROUPS : SPG_YIELD_KMV_HASH;
           }
         } else if (rb_type(pg_value_key) == T_ARRAY) {
           if (SYMBOL_P(pg_value_value)) {
-            type = SPG_YIELD_MKV_HASH;
+            type = pg_type == spg_sym_hash_groups ? SPG_YIELD_MKV_HASH_GROUPS : SPG_YIELD_MKV_HASH;
           } else if (rb_type(pg_value_value) == T_ARRAY) {
-            type = SPG_YIELD_MKMV_HASH;
+            type = pg_type == spg_sym_hash_groups ? SPG_YIELD_MKMV_HASH_GROUPS : SPG_YIELD_MKMV_HASH;
           }
         }
       } else if (pg_type == spg_sym_model && rb_type(pg_value) == T_CLASS) {
@@ -515,54 +679,114 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
       } 
       break;
     case SPG_YIELD_KV_HASH:
+    case SPG_YIELD_KV_HASH_GROUPS:
       /* Hash with single key and single value */
       {
         VALUE k, v;
         h = rb_hash_new();
         k = spg__field_id(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_id(rb_ary_entry(pg_value, 1), colsyms, nfields);
-        for(i=0; i<ntuples; i++) {
-          rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert ENC_INDEX), spg__col_value(self, res, i, v, colconvert ENC_INDEX));
-        } 
+        if(type == SPG_YIELD_KV_HASH) {
+          for(i=0; i<ntuples; i++) {
+            rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert ENC_INDEX), spg__col_value(self, res, i, v, colconvert ENC_INDEX));
+          } 
+        } else {
+          VALUE kv, vv, a;
+          for(i=0; i<ntuples; i++) {
+            kv = spg__col_value(self, res, i, k, colconvert ENC_INDEX);
+            vv = spg__col_value(self, res, i, v, colconvert ENC_INDEX);
+            a = rb_hash_lookup(h, kv);
+            if(!RTEST(a)) {
+              rb_hash_aset(h, kv, rb_ary_new3(1, vv));
+            } else {
+              rb_ary_push(a, vv);
+            }
+          } 
+        }
         rb_yield(h);
       }
       break;
     case SPG_YIELD_MKV_HASH:
+    case SPG_YIELD_MKV_HASH_GROUPS:
       /* Hash with array of keys and single value */
       {
         VALUE k, v;
         h = rb_hash_new();
         k = spg__field_ids(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_id(rb_ary_entry(pg_value, 1), colsyms, nfields);
-        for(i=0; i<ntuples; i++) {
-          rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX), spg__col_value(self, res, i, v, colconvert ENC_INDEX));
-        } 
+        if(type == SPG_YIELD_MKV_HASH) {
+          for(i=0; i<ntuples; i++) {
+            rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX), spg__col_value(self, res, i, v, colconvert ENC_INDEX));
+          } 
+        } else {
+          VALUE kv, vv, a;
+          for(i=0; i<ntuples; i++) {
+            kv = spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX);
+            vv = spg__col_value(self, res, i, v, colconvert ENC_INDEX);
+            a = rb_hash_lookup(h, kv);
+            if(!RTEST(a)) {
+              rb_hash_aset(h, kv, rb_ary_new3(1, vv));
+            } else {
+              rb_ary_push(a, vv);
+            }
+          } 
+        }
         rb_yield(h);
       }
       break;
     case SPG_YIELD_KMV_HASH:
+    case SPG_YIELD_KMV_HASH_GROUPS:
       /* Hash with single keys and array of values */
       {
         VALUE k, v;
         h = rb_hash_new();
         k = spg__field_id(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_ids(rb_ary_entry(pg_value, 1), colsyms, nfields);
-        for(i=0; i<ntuples; i++) {
-          rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert ENC_INDEX), spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX));
-        } 
+        if(type == SPG_YIELD_KMV_HASH) {
+          for(i=0; i<ntuples; i++) {
+            rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert ENC_INDEX), spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX));
+          } 
+        } else {
+          VALUE kv, vv, a;
+          for(i=0; i<ntuples; i++) {
+            kv = spg__col_value(self, res, i, k, colconvert ENC_INDEX);
+            vv = spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX);
+            a = rb_hash_lookup(h, kv);
+            if(!RTEST(a)) {
+              rb_hash_aset(h, kv, rb_ary_new3(1, vv));
+            } else {
+              rb_ary_push(a, vv);
+            }
+          } 
+        }
         rb_yield(h);
       }
       break;
     case SPG_YIELD_MKMV_HASH:
+    case SPG_YIELD_MKMV_HASH_GROUPS:
       /* Hash with array of keys and array of values */
       {
         VALUE k, v;
         h = rb_hash_new();
         k = spg__field_ids(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_ids(rb_ary_entry(pg_value, 1), colsyms, nfields);
-        for(i=0; i<ntuples; i++) {
-          rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX), spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX));
-        } 
+        if(type == SPG_YIELD_MKMV_HASH) {
+          for(i=0; i<ntuples; i++) {
+            rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX), spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX));
+          } 
+        } else {
+          VALUE kv, vv, a;
+          for(i=0; i<ntuples; i++) {
+            kv = spg__col_values(self, k, colsyms, nfields, res, i, colconvert ENC_INDEX);
+            vv = spg__col_values(self, v, colsyms, nfields, res, i, colconvert ENC_INDEX);
+            a = rb_hash_lookup(h, kv);
+            if(!RTEST(a)) {
+              rb_hash_aset(h, kv, rb_ary_new3(1, vv));
+            } else {
+              rb_ary_push(a, vv);
+            }
+          } 
+        }
         rb_yield(h);
       }
       break;
@@ -583,6 +807,199 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
   return self;
 }
 
+static VALUE spg_supports_streaming_p(VALUE self) {
+  return
+#if HAVE_PQSETROWPROCESSOR
+  Qtrue;
+#else
+  Qfalse;
+#endif
+}
+
+#if HAVE_PQSETROWPROCESSOR
+static VALUE spg__rp_value(VALUE self, PGresult* res, const PGdataValue* dvs, int j, VALUE* colconvert
+#ifdef SPG_ENCODING
+, int enc_index
+#endif
+) {
+  const char *v;
+  PGdataValue dv = dvs[j];
+  VALUE rv;
+  size_t l;
+  int len = dv.len;
+
+  if(len < 0) {
+    rv = Qnil;
+  } else {
+    v = dv.value;
+
+    switch(PQftype(res, j)) {
+      case 16: /* boolean */
+        rv = *v == 't' ? Qtrue : Qfalse;
+        break;
+      case 17: /* bytea */
+        v = PQunescapeBytea((unsigned char*)v, &l);
+        rv = rb_funcall(spg_Blob, spg_id_new, 1, rb_str_new(v, l));
+        PQfreemem((char *)v);
+        break;
+      case 20: /* integer */
+      case 21:
+      case 22:
+      case 23:
+      case 26:
+        rv = rb_str2inum(rb_str_new(v, len), 10);
+        break;
+      case 700: /* float */
+      case 701:
+        if (strncmp("NaN", v, 3) == 0) {
+          rv = spg_nan;
+        } else if (strncmp("Infinity", v, 8) == 0) {
+          rv = spg_pos_inf;
+        } else if (strncmp("-Infinity", v, 9) == 0) {
+          rv = spg_neg_inf;
+        } else {
+          rv = rb_float_new(rb_str_to_dbl(rb_str_new(v, len), Qfalse));
+        }
+        break;
+      case 790: /* numeric */
+      case 1700:
+        rv = rb_funcall(spg_BigDecimal, spg_id_new, 1, rb_str_new(v, len));
+        break;
+      case 1082: /* date */
+        rv = rb_str_new(v, len);
+        rv = spg_date(StringValuePtr(rv));
+        break;
+      case 1083: /* time */
+      case 1266:
+        rv = rb_str_new(v, len);
+        rv = spg_time(StringValuePtr(rv));
+        break;
+      case 1114: /* timestamp */
+      case 1184:
+        rv = rb_str_new(v, len);
+        rv = spg_timestamp(StringValuePtr(rv), self);
+        break;
+      case 18: /* char */
+      case 25: /* text */
+      case 1043: /* varchar*/
+        rv = rb_tainted_str_new(v, len);
+#ifdef SPG_ENCODING
+        rb_enc_associate_index(rv, enc_index);
+#endif
+        break;
+      default:
+        rv = rb_tainted_str_new(v, len);
+#ifdef SPG_ENCODING
+        rb_enc_associate_index(rv, enc_index);
+#endif
+        if (colconvert[j] != Qnil) {
+          rv = rb_funcall(colconvert[j], spg_id_call, 1, rv); 
+        }
+    }
+  }
+  return rv;
+}
+
+static int spg_row_processor(PGresult *res, const PGdataValue *columns, const char **errmsgp, void *param) {
+  long nfields;
+  struct spg_row_proc_info *info;
+  info = (struct spg_row_proc_info *)param;
+  VALUE *colsyms = info->colsyms;
+  VALUE *colconvert = info->colconvert;
+  VALUE self = info->dataset;
+
+  switch (PQresultStatus(res))
+  {
+    case PGRES_TUPLES_OK:
+    case PGRES_COPY_OUT:
+    case PGRES_COPY_IN:
+#ifdef HAVE_CONST_PGRES_COPY_BOTH
+    case PGRES_COPY_BOTH:
+#endif
+    case PGRES_EMPTY_QUERY:
+    case PGRES_COMMAND_OK:
+      break;
+    case PGRES_BAD_RESPONSE:
+    case PGRES_FATAL_ERROR:
+    case PGRES_NONFATAL_ERROR:
+      rb_raise(spg_PGError, "error while streaming results");
+    default:
+      rb_raise(spg_PGError, "unexpected result status while streaming results");
+  }
+
+  nfields = PQnfields(res);
+  if(columns == NULL) {
+    spg_set_column_info(self, res, colsyms, colconvert);
+    rb_ivar_set(self, spg_id_columns, rb_ary_new4(nfields, colsyms));
+  } else {
+    long j;
+    VALUE h, m;
+    h = rb_hash_new();
+
+    for(j=0; j<nfields; j++) {
+      rb_hash_aset(h, colsyms[j], spg__rp_value(self, res, columns, j, colconvert
+#ifdef SPG_ENCODING
+                   , info->enc_index
+#endif
+      ));
+    }
+
+    /* optimize_model_load used, return model instance */
+    if ((m = info->model)) {
+      m = rb_obj_alloc(m);
+      rb_ivar_set(m, spg_id_values, h);
+      h = m;
+    }
+
+    rb_funcall(info->block, spg_id_call, 1, h);
+  }
+  return 1;
+}
+
+static VALUE spg_unset_row_processor(VALUE rconn) {
+  PGconn *conn;
+  Data_Get_Struct(rconn, PGconn, conn);
+  if ((PQskipResult(conn)) != NULL) {
+    /* Results remaining when row processor finished,
+     * either because an exception was raised or the iterator
+     * exited early, so skip all remaining rows. */
+    while(PQgetResult(conn) != NULL) {
+      /* Use a separate while loop as PQgetResult is faster than
+       * PQskipResult. */
+    }
+  }
+  PQsetRowProcessor(conn, NULL, NULL);
+  return Qnil;
+}
+
+static VALUE spg_with_row_processor(VALUE self, VALUE rconn, VALUE dataset, VALUE block) {
+  struct spg_row_proc_info info;
+  PGconn *conn;
+  Data_Get_Struct(rconn, PGconn, conn);
+  bzero(&info, sizeof(info));
+
+  info.dataset = dataset;
+  info.block = block;
+  info.model = 0;
+#if SPG_ENCODING
+  info.enc_index = enc_get_index(rconn);
+#endif
+
+  /* Abuse local variable, detect if optimize_model_load used */
+  block = rb_funcall(dataset, spg_id_opts, 0);
+  if (rb_type(block) == T_HASH && rb_hash_aref(block, spg_sym__sequel_pg_type) == spg_sym_model) {
+    block = rb_hash_aref(block, spg_sym__sequel_pg_value);
+    if (rb_type(block) == T_CLASS) {
+      info.model = block;
+    }
+  }
+
+  PQsetRowProcessor(conn, spg_row_processor, (void*)&info);
+  rb_ensure(rb_yield, Qnil, spg_unset_row_processor, rconn);
+  return Qnil;
+}
+#endif
+
 void Init_sequel_pg(void) {
   VALUE c, spg_Postgres;
   ID cg;
@@ -601,6 +1018,8 @@ void Init_sequel_pg(void) {
   spg_id_utc_offset = rb_intern("utc_offset");
   spg_id_localtime = rb_intern("localtime");
   spg_id_new_offset = rb_intern("new_offset");
+  spg_id_convert_infinite_timestamps = rb_intern("convert_infinite_timestamps");
+  spg_id_infinite_timestamp_value = rb_intern("infinite_timestamp_value");
 
   spg_id_call = rb_intern("call");
   spg_id_get = rb_intern("[]");
@@ -620,6 +1039,7 @@ void Init_sequel_pg(void) {
   spg_sym_first = ID2SYM(rb_intern("first"));
   spg_sym_array = ID2SYM(rb_intern("array"));
   spg_sym_hash = ID2SYM(rb_intern("hash"));
+  spg_sym_hash_groups = ID2SYM(rb_intern("hash_groups"));
   spg_sym_model = ID2SYM(rb_intern("model"));
   spg_sym__sequel_pg_type = ID2SYM(rb_intern("_sequel_pg_type"));
   spg_sym__sequel_pg_value = ID2SYM(rb_intern("_sequel_pg_value"));
@@ -630,15 +1050,39 @@ void Init_sequel_pg(void) {
   spg_BigDecimal = rb_funcall(rb_cObject, cg, 1, rb_str_new2("BigDecimal")); 
   spg_Date = rb_funcall(rb_cObject, cg, 1, rb_str_new2("Date")); 
   spg_Postgres = rb_funcall(spg_Sequel, cg, 1, rb_str_new2("Postgres"));
+  spg_PGError = rb_funcall(rb_cObject, cg, 1, rb_str_new2("PGError"));
+
+  spg_nan = rb_eval_string("0.0/0.0");
+  spg_pos_inf = rb_eval_string("1.0/0.0");
+  spg_neg_inf = rb_eval_string("-1.0/0.0");
 
   rb_global_variable(&spg_Sequel);
   rb_global_variable(&spg_Blob);
   rb_global_variable(&spg_BigDecimal);
   rb_global_variable(&spg_Date);
+  rb_global_variable(&spg_SQLTime);
+  rb_global_variable(&spg_Postgres);
+  rb_global_variable(&spg_nan);
+  rb_global_variable(&spg_pos_inf);
+  rb_global_variable(&spg_neg_inf);
+
+  /* Check for 1.8-1.9.2 stdlib date that needs Rational for usec accuracy */
+  if (rb_eval_string("Date.today.instance_variable_get(:@ajd)") != Qnil) {
+    spg_id_Rational = rb_intern("Rational");
+  }
 
   c = rb_funcall(spg_Postgres, cg, 1, rb_str_new2("Dataset"));
   rb_define_private_method(c, "yield_hash_rows", spg_yield_hash_rows, 2);
   rb_define_private_method(c, "fetch_rows_set_cols", spg_fetch_rows_set_cols, 1);
 
+  rb_define_singleton_method(spg_Postgres, "supports_streaming?", spg_supports_streaming_p, 0);
+
+#if HAVE_PQSETROWPROCESSOR
+  c = rb_funcall(spg_Postgres, cg, 1, rb_str_new2("Database"));
+  rb_define_private_method(c, "with_row_processor", spg_with_row_processor, 3);
+#endif
+
+  rb_define_singleton_method(spg_Postgres, "parse_pg_array", parse_pg_array, 2);
+
   rb_require("sequel_pg/sequel_pg");
 }
